@@ -116,12 +116,15 @@ typedef struct {
         gpointer completion_data;
         SoupHTTP2IOState state;
         GError *error;
+        uint32_t http2_error;
         gboolean paused;
         guint32 stream_id;
         gboolean can_be_restarted;
+        gboolean expect_continue;
 } SoupHTTP2MessageData;
 
 static void soup_client_message_io_http2_finished (SoupClientMessageIO *iface, SoupMessage *msg);
+static ssize_t on_data_source_read_callback (nghttp2_session *session, int32_t stream_id, uint8_t *buf, size_t length, uint32_t *data_flags, nghttp2_data_source *source, void *user_data);
 
 static void
 NGCHECK (int return_code)
@@ -168,6 +171,22 @@ frame_type_to_string (nghttp2_frame_type type)
         }
 }
 
+static const char *
+headers_category_to_string (nghttp2_headers_category catergory)
+{
+        switch (catergory) {
+        case NGHTTP2_HCAT_REQUEST:
+                return "REQUEST";
+        case NGHTTP2_HCAT_RESPONSE:
+                return "RESPONSE";
+        case NGHTTP2_HCAT_PUSH_RESPONSE:
+                return "PUSH_RESPONSE";
+        case NGHTTP2_HCAT_HEADERS:
+                return "HEADERS";
+        }
+        g_assert_not_reached ();
+}
+
 static const char *
 state_to_string (SoupHTTP2IOState state)
 {
@@ -249,6 +268,20 @@ set_error_for_data (SoupHTTP2MessageData *data,
                 g_error_free (error);
 }
 
+static void
+set_http2_error_for_data (SoupHTTP2MessageData *data,
+                          uint32_t              error_code)
+{
+        h2_debug (data->io, data, "[SESSION] Error: %s", nghttp2_http2_strerror (error_code));
+
+        if (data->error)
+                return;
+
+        data->http2_error = error_code;
+        data->error = g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED,
+                                   "HTTP/2 Error: %s", nghttp2_http2_strerror (error_code));
+}
+
 static void
 set_io_error (SoupClientMessageIOHTTP2 *io,
               GError                   *error)
@@ -259,6 +292,11 @@ set_io_error (SoupClientMessageIOHTTP2 *io,
                 io->error = error;
         else
                 g_error_free (error);
+
+        if (io->close_task && !io->goaway_sent) {
+                g_task_return_boolean (io->close_task, TRUE);
+                g_clear_object (&io->close_task);
+        }
 }
 
 static void
@@ -285,6 +323,22 @@ advance_state_from (SoupHTTP2MessageData *data,
         data->state = to;
 }
 
+static gboolean
+soup_http2_message_data_can_be_restarted (SoupHTTP2MessageData *data,
+                                          GError               *error)
+{
+        if (data->can_be_restarted)
+                return TRUE;
+
+        return data->state < STATE_READ_DATA_START &&
+                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_TIMED_OUT) &&
+                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK) &&
+                !g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED) &&
+                error->domain != G_TLS_ERROR &&
+                data->http2_error == NGHTTP2_NO_ERROR &&
+                SOUP_METHOD_IS_IDEMPOTENT (soup_message_get_method (data->msg));
+}
+
 static void
 soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
 {
@@ -296,6 +350,7 @@ soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
         if (g_cancellable_set_error_if_cancelled (g_task_get_cancellable (task), &error)) {
                 io->pending_io_messages = g_list_remove (io->pending_io_messages, data);
                 data->task = NULL;
+                soup_client_message_io_http2_finished ((SoupClientMessageIO *)io, msg);
                 g_task_return_error (task, error);
                 g_object_unref (task);
                 return;
@@ -310,7 +365,7 @@ soup_http2_message_data_check_status (SoupHTTP2MessageData *data)
         if (data->error) {
                 GError *error = g_steal_pointer (&data->error);
 
-                if (data->can_be_restarted)
+                if (soup_http2_message_data_can_be_restarted (data, error))
                         data->item->state = SOUP_MESSAGE_RESTARTING;
                 else
                         soup_message_set_metrics_timestamp (data->msg, SOUP_MESSAGE_METRICS_RESPONSE_END);
@@ -347,7 +402,7 @@ io_write (SoupClientMessageIOHTTP2 *io,
 
         if (io->write_buffer == NULL) {
                 io->written_bytes = 0;
-                g_assert (io->in_callback == 0);
+                g_warn_if_fail (io->in_callback == 0);
                 io->write_buffer_size = nghttp2_session_mem_send (io->session, (const guint8**)&io->write_buffer);
                 NGCHECK (io->write_buffer_size);
                 if (io->write_buffer_size == 0) {
@@ -379,7 +434,7 @@ io_write_ready (GObject                  *stream,
                 return G_SOURCE_REMOVE;
         }
 
-        while (nghttp2_session_want_write (io->session) && !error)
+        while (!error && nghttp2_session_want_write (io->session))
                 io_write (io, FALSE, NULL, &error);
 
         if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_WOULD_BLOCK)) {
@@ -407,7 +462,7 @@ io_try_write (SoupClientMessageIOHTTP2 *io,
                 if (blocking || !nghttp2_session_want_write (io->session))
                         return;
         } else {
-                while (nghttp2_session_want_write (io->session) && !error)
+                while (!error && nghttp2_session_want_write (io->session))
                         io_write (io, blocking, NULL, &error);
         }
 
@@ -415,6 +470,8 @@ io_try_write (SoupClientMessageIOHTTP2 *io,
                 g_clear_error (&error);
                 io->write_source = g_pollable_output_stream_create_source (G_POLLABLE_OUTPUT_STREAM (io->ostream), NULL);
                 g_source_set_name (io->write_source, "Soup HTTP/2 write source");
+                /* Give write more priority than read */
+                g_source_set_priority (io->write_source, G_PRIORITY_DEFAULT - 1);
                 g_source_set_callback (io->write_source, (GSourceFunc)io_write_ready, io, NULL);
                 g_source_attach (io->write_source, g_main_context_get_thread_default ());
         }
@@ -433,14 +490,24 @@ io_read (SoupClientMessageIOHTTP2  *io,
         gssize read;
         int ret;
 
+        /* Always try to write before read, in case there's a pending reset stream after an error. */
+        io_try_write (io, blocking);
+
         if ((read = g_pollable_stream_read (io->istream, buffer, sizeof (buffer),
                                             blocking, cancellable, error)) < 0)
             return FALSE;
 
-        g_assert (io->in_callback == 0);
+        if (read == 0) {
+                g_set_error_literal (error, G_IO_ERROR,
+                                     G_IO_ERROR_PARTIAL_INPUT,
+                                     _("Connection terminated unexpectedly"));
+                return FALSE;
+        }
+
+        g_warn_if_fail (io->in_callback == 0);
         ret = nghttp2_session_mem_recv (io->session, buffer, read);
         NGCHECK (ret);
-        return ret != 0;
+        return ret > 0;
 }
 
 static gboolean
@@ -461,7 +528,7 @@ io_read_ready (GObject                  *stream,
         if (io->conn)
                 soup_connection_set_in_use (io->conn, TRUE);
 
-        while (nghttp2_session_want_read (io->session) && progress) {
+        while (progress && nghttp2_session_want_read (io->session)) {
                 progress = io_read (io, FALSE, NULL, &error);
                 if (progress) {
                         g_list_foreach (io->pending_io_messages,
@@ -477,6 +544,8 @@ io_read_ready (GObject                  *stream,
                 return G_SOURCE_CONTINUE;
         }
 
+        io->is_shutdown = TRUE;
+
         if (error) {
                 set_io_error (io, error);
                 g_list_foreach (io->pending_io_messages,
@@ -484,8 +553,6 @@ io_read_ready (GObject                  *stream,
                                 NULL);
         }
 
-        io->is_shutdown = TRUE;
-
         g_clear_pointer (&io->read_source, g_source_unref);
         if (io->conn)
                 soup_connection_set_in_use (io->conn, FALSE);
@@ -571,6 +638,7 @@ on_header_callback (nghttp2_session     *session,
 
 static GError *
 memory_stream_need_more_data_callback (SoupBodyInputStreamHttp2 *stream,
+                                       gboolean                  blocking,
                                        GCancellable             *cancellable,
                                        gpointer                  user_data)
 {
@@ -578,7 +646,7 @@ memory_stream_need_more_data_callback (SoupBodyInputStreamHttp2 *stream,
         GError *error = NULL;
 
         if (nghttp2_session_want_read (data->io->session))
-                io_read (data->io, TRUE, cancellable, &error);
+                io_read (data->io, blocking, cancellable, &error);
 
         return error;
 }
@@ -599,7 +667,7 @@ on_begin_frame_callback (nghttp2_session        *session,
 
         switch (hd->type) {
         case NGHTTP2_HEADERS:
-                if (data->state < STATE_READ_HEADERS) {
+                if (data->state == STATE_WRITE_DONE) {
                         soup_message_set_metrics_timestamp (data->item->msg, SOUP_MESSAGE_METRICS_RESPONSE_START);
                         advance_state_from (data, STATE_WRITE_DONE, STATE_READ_HEADERS);
                 }
@@ -634,6 +702,9 @@ handle_goaway (SoupClientMessageIOHTTP2 *io,
         GHashTableIter iter;
         SoupHTTP2MessageData *data;
 
+        if (last_stream_id == G_MAXINT32)
+                return;
+
         g_hash_table_iter_init (&iter, io->messages);
         while (g_hash_table_iter_next (&iter, NULL, (gpointer*)&data)) {
                 /* If there is no error it is a graceful shutdown and
@@ -641,8 +712,7 @@ handle_goaway (SoupClientMessageIOHTTP2 *io,
                 if ((error_code == 0 && (int32_t)data->stream_id > last_stream_id) ||
                      data->state < STATE_READ_DONE) {
                         /* TODO: We can restart unfinished messages */
-                        set_error_for_data (data, g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED,
-                                            "HTTP/2 Error: %s", nghttp2_http2_strerror (error_code)));
+                        set_http2_error_for_data (data, error_code);
                 }
         }
 }
@@ -691,47 +761,73 @@ on_frame_recv_callback (nghttp2_session     *session,
         }
 
         switch (frame->hd.type) {
-        case NGHTTP2_HEADERS:
+        case NGHTTP2_HEADERS: {
+                guint status = soup_message_get_status (data->msg);
+
                 if (data->metrics)
                         data->metrics->response_header_bytes_received += frame->hd.length + FRAME_HEADER_SIZE;
 
-                if (frame->headers.cat == NGHTTP2_HCAT_RESPONSE && frame->hd.flags & NGHTTP2_FLAG_END_HEADERS) {
-                        h2_debug (io, data, "[HEADERS] status %u", soup_message_get_status (data->msg));
-                        if (SOUP_STATUS_IS_INFORMATIONAL (soup_message_get_status (data->msg))) {
-                                soup_message_got_informational (data->msg);
-                                soup_message_cleanup_response (data->msg);
-                                advance_state_from (data, STATE_READ_HEADERS, STATE_READ_DONE);
+                h2_debug (io, data, "[HEADERS] category=%s status=%u",
+                          headers_category_to_string (frame->headers.cat), status);
+                switch (frame->headers.cat) {
+                case NGHTTP2_HCAT_HEADERS:
+                        if (!(frame->hd.flags & NGHTTP2_FLAG_END_HEADERS)) {
                                 io->in_callback--;
                                 return 0;
                         }
+                        break;
+                case NGHTTP2_HCAT_RESPONSE:
+                        if (SOUP_STATUS_IS_INFORMATIONAL (status)) {
+                                if (data->expect_continue && status == SOUP_STATUS_CONTINUE) {
+                                        nghttp2_data_provider data_provider;
 
-                        if (soup_message_get_status (data->msg) == SOUP_STATUS_NO_CONTENT ||
-                            frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
-                                h2_debug (io, data, "Stream done");
-                                advance_state_from (data, STATE_READ_HEADERS, STATE_READ_DATA);
+                                        data_provider.source.ptr = soup_message_get_request_body_stream (data->msg);
+                                        data_provider.read_callback = on_data_source_read_callback;
+                                        nghttp2_submit_data (io->session, NGHTTP2_FLAG_END_STREAM, frame->hd.stream_id, &data_provider);
+                                        io_try_write (io, !data->item->async);
+                                }
+
+                                soup_message_got_informational (data->msg);
+                                soup_message_cleanup_response (data->msg);
+                                io->in_callback--;
+                                return 0;
                         }
-                        soup_message_got_headers (data->msg);
+                        break;
+                case NGHTTP2_HCAT_PUSH_RESPONSE:
+                        g_warn_if_reached ();
+                        break;
+                default:
+                        g_assert_not_reached ();
                 }
+
+                if (soup_message_get_status (data->msg) == SOUP_STATUS_NO_CONTENT || frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
+                        h2_debug (io, data, "Stream done");
+                        advance_state_from (data, STATE_READ_HEADERS, STATE_READ_DATA);
+                }
+                soup_message_got_headers (data->msg);
+
                 break;
+        }
         case NGHTTP2_DATA:
                 if (data->metrics)
                         data->metrics->response_body_bytes_received += frame->data.hd.length + FRAME_HEADER_SIZE;
-                if (frame->hd.flags & NGHTTP2_FLAG_END_STREAM && data->body_istream) {
-                        soup_body_input_stream_http2_complete (SOUP_BODY_INPUT_STREAM_HTTP2 (data->body_istream));
-                        if (data->state == STATE_READ_DATA_START) {
-                                io_try_sniff_content (data, FALSE, data->item->cancellable);
-                                if (data->state == STATE_READ_DATA && data->item->async)
-                                        soup_http2_message_data_check_status (data);
+                if (frame->hd.flags & NGHTTP2_FLAG_END_STREAM) {
+                        if (data->body_istream) {
+                                soup_body_input_stream_http2_complete (SOUP_BODY_INPUT_STREAM_HTTP2 (data->body_istream));
+                                if (data->state == STATE_READ_DATA_START) {
+                                        io_try_sniff_content (data, FALSE, data->item->cancellable);
+                                        if (data->state == STATE_READ_DATA && data->item->async)
+                                                soup_http2_message_data_check_status (data);
+                                }
                         }
+                } else {
+                        /* Try to write after every data frame, since nghttp2 might need to send a window update. */
+                        io_try_write (io, !data->item->async);
                 }
-                /* Try to write after every data frame, since nghttp2 might need to send a window update. */
-                io_try_write (io, !data->item->async);
                 break;
         case NGHTTP2_RST_STREAM:
-                if (frame->rst_stream.error_code != NGHTTP2_NO_ERROR) {
-                        set_error_for_data (data, g_error_new_literal (G_IO_ERROR, G_IO_ERROR_FAILED,
-                                                                       nghttp2_http2_strerror (frame->rst_stream.error_code)));
-                }
+                if (frame->rst_stream.error_code != NGHTTP2_NO_ERROR)
+                        set_http2_error_for_data (data, frame->rst_stream.error_code);
                 break;
         case NGHTTP2_WINDOW_UPDATE:
                 h2_debug (io, data, "[RECV] WINDOW_UPDATE: increment=%d, total=%d", frame->window_update.window_size_increment,
@@ -826,7 +922,8 @@ on_frame_send_callback (nghttp2_session     *session,
         switch (frame->hd.type) {
         case NGHTTP2_HEADERS:
                 g_assert (data);
-                h2_debug (io, data, "[SEND] [HEADERS] finished=%d",
+                h2_debug (io, data, "[SEND] [HEADERS] category=%s finished=%d",
+                          headers_category_to_string (frame->headers.cat),
                           (frame->hd.flags & NGHTTP2_FLAG_END_HEADERS) ? 1 : 0);
 
                 if (data->metrics)
@@ -913,8 +1010,21 @@ on_stream_close_callback (nghttp2_session *session,
 
         data->io->in_callback++;
 
-        if (error_code == NGHTTP2_REFUSED_STREAM && data->state < STATE_READ_DATA)
+        switch (error_code) {
+        case NGHTTP2_NO_ERROR:
+                break;
+        case NGHTTP2_REFUSED_STREAM:
+                if (data->state < STATE_READ_DATA_START)
+                        data->can_be_restarted = TRUE;
+                break;
+        case NGHTTP2_HTTP_1_1_REQUIRED:
+                soup_message_set_force_http_version (data->item->msg, SOUP_HTTP_1_1);
                 data->can_be_restarted = TRUE;
+                break;
+        default:
+                set_http2_error_for_data (data, error_code);
+                break;
+        }
 
         data->io->in_callback--;
         return 0;
@@ -944,8 +1054,6 @@ on_data_read (GInputStream *source,
         GError *error = NULL;
         gssize read = g_input_stream_read_finish (source, res, &error);
 
-        h2_debug (data->io, data, "[SEND_BODY] Read %zd", read);
-
         /* This operation may have outlived the message data in which
            case this will have been cancelled. */
         if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
@@ -953,6 +1061,8 @@ on_data_read (GInputStream *source,
                 return;
         }
 
+        h2_debug (data->io, data, "[SEND_BODY] Read %zd", read);
+
         if (read < 0) {
                 g_byte_array_set_size (data->data_source_buffer, 0);
                 data->data_source_error = g_steal_pointer (&error);
@@ -993,6 +1103,31 @@ on_data_source_read_callback (nghttp2_session     *session,
         SoupHTTP2MessageData *data = nghttp2_session_get_stream_user_data (session, stream_id);
         data->io->in_callback++;
 
+        if (!data->item->async) {
+                gssize read;
+                GError *error = NULL;
+
+                read = g_input_stream_read (source->ptr, buf, length, data->item->cancellable, &error);
+                if (read) {
+                        h2_debug (data->io, data, "[SEND_BODY] Read %zd", read);
+                        log_request_data (data, buf, read);
+                }
+
+                if (read < 0) {
+                        set_error_for_data (data, g_steal_pointer (&error));
+                        data->io->in_callback--;
+                        return NGHTTP2_ERR_TEMPORAL_CALLBACK_FAILURE;
+                }
+
+                if (read == 0) {
+                        h2_debug (data->io, data, "[SEND_BODY] EOF");
+                        *data_flags |= NGHTTP2_DATA_FLAG_EOF;
+                }
+
+                data->io->in_callback--;
+                return read;
+        }
+
         /* We support pollable streams in the best case because they
          * should perform better with one fewer copy of each buffer and no threading. */
         if (G_IS_POLLABLE_INPUT_STREAM (source->ptr) && g_pollable_input_stream_can_poll (G_POLLABLE_INPUT_STREAM (source->ptr))) {
@@ -1267,13 +1402,18 @@ send_message_request (SoupMessage          *msg,
         nghttp2_priority_spec priority_spec;
         nghttp2_priority_spec_init (&priority_spec, 0, message_priority_to_weight (msg), 0);
 
-        nghttp2_data_provider data_provider;
-        if (body_stream) {
-                data_provider.source.ptr = body_stream;
-                data_provider.read_callback = on_data_source_read_callback;
+        int32_t stream_id;
+        if (body_stream && soup_message_headers_get_expectations (soup_message_get_request_headers (msg)) & SOUP_EXPECTATION_CONTINUE) {
+                data->expect_continue = TRUE;
+                stream_id = nghttp2_submit_headers (io->session, 0, -1, &priority_spec, (const nghttp2_nv *)headers->data, headers->len, data);
+        } else {
+                nghttp2_data_provider data_provider;
+                if (body_stream) {
+                        data_provider.source.ptr = body_stream;
+                        data_provider.read_callback = on_data_source_read_callback;
+                }
+                stream_id = nghttp2_submit_request (io->session, &priority_spec, (const nghttp2_nv *)headers->data, headers->len, body_stream ? &data_provider : NULL, data);
         }
-
-        int32_t stream_id = nghttp2_submit_request (io->session, &priority_spec, (const nghttp2_nv *)headers->data, headers->len, body_stream ? &data_provider : NULL, data);
         if (stream_id == NGHTTP2_ERR_STREAM_ID_NOT_AVAILABLE) {
                 set_error_for_data (data,
                                     g_error_new_literal (G_IO_ERROR, G_IO_ERROR_FAILED,
@@ -1319,6 +1459,7 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
 	SoupMessageIOCompletionFn completion_cb;
 	gpointer completion_data;
         SoupMessageIOCompletion completion;
+        gboolean is_closed;
 
         data = get_data_for_message (io, msg);
 
@@ -1331,9 +1472,10 @@ soup_client_message_io_http2_finished (SoupClientMessageIO *iface,
 
 	g_object_ref (msg);
 
+        is_closed = nghttp2_session_get_stream_user_data (io->session, data->stream_id) == NULL;
         nghttp2_session_set_stream_user_data (io->session, data->stream_id, NULL);
 
-        if (!io->is_shutdown) {
+        if (!io->is_shutdown && !is_closed) {
                 NGCHECK (nghttp2_submit_rst_stream (io->session, NGHTTP2_FLAG_NONE, data->stream_id,
                                                     completion == SOUP_MESSAGE_IO_COMPLETE ? NGHTTP2_NO_ERROR : NGHTTP2_CANCEL));
                 soup_http2_message_data_close (data);
@@ -1528,11 +1670,13 @@ io_run_until (SoupClientMessageIOHTTP2 *io,
                 progress = io_run (data, cancellable, &my_error);
 
         if (my_error) {
-                g_propagate_error (error, my_error);
-                g_object_unref (msg);
-                return FALSE;
+                io->is_shutdown = TRUE;
+                set_io_error (io, my_error);
         }
 
+        if (io->error && !data->error)
+                data->error = g_error_copy (io->error);
+
 	if (data->error) {
                 g_propagate_error (error, g_steal_pointer (&data->error));
 		g_object_unref (msg);
@@ -1561,12 +1705,13 @@ soup_client_message_io_http2_run_until_read (SoupClientMessageIO  *iface,
 {
         SoupClientMessageIOHTTP2 *io = (SoupClientMessageIOHTTP2 *)iface;
         SoupHTTP2MessageData *data = get_data_for_message (io, msg);
+        GError *my_error = NULL;
 
-        if (io_run_until (io, msg, STATE_READ_DATA, cancellable, error))
+        if (io_run_until (io, msg, STATE_READ_DATA, cancellable, &my_error))
                 return TRUE;
 
         if (get_io_data (msg) == io) {
-                if (data->can_be_restarted)
+                if (soup_http2_message_data_can_be_restarted (data, my_error))
                         data->item->state = SOUP_MESSAGE_RESTARTING;
                 else
                         soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_RESPONSE_END);
@@ -1574,6 +1719,8 @@ soup_client_message_io_http2_run_until_read (SoupClientMessageIO  *iface,
                 soup_client_message_io_http2_finished (iface, msg);
         }
 
+        g_propagate_error (error, my_error);
+
         return FALSE;
 }
 
@@ -1756,6 +1903,7 @@ soup_client_message_io_http2_new (SoupConnection *conn)
 
         io->read_source = g_pollable_input_stream_create_source (G_POLLABLE_INPUT_STREAM (io->istream), NULL);
         g_source_set_name (io->read_source, "Soup HTTP/2 read source");
+        g_source_set_priority (io->read_source, G_PRIORITY_DEFAULT);
         g_source_set_callback (io->read_source, (GSourceFunc)io_read_ready, io, NULL);
         g_source_attach (io->read_source, g_main_context_get_thread_default ());
 
