@@ -217,12 +217,8 @@ do_cancellation_test (Test *test, gconstpointer data)
         soup_session_send_and_read_async (test->session, msg, G_PRIORITY_DEFAULT, cancellable,
                                           (GAsyncReadyCallback)on_send_and_read_cancelled_complete, &done);
 
-        /* Just iterate until a partial read is happening */
-        for (guint i = 100000; i; i--)
-                g_main_context_iteration (async_context, FALSE);
-
-        /* Then cancel everything */
-        g_cancellable_cancel (cancellable);
+        /* Cancel right after getting the headers */
+        g_signal_connect_swapped (msg, "got-headers", G_CALLBACK (g_cancellable_cancel), cancellable);
 
         while (!done)
                 g_main_context_iteration (async_context, FALSE);
@@ -1013,6 +1009,39 @@ do_timeout_test (Test *test, gconstpointer data)
                 g_main_context_iteration (NULL, FALSE);
 }
 
+static void
+message_restarted (SoupMessage *msg,
+                   gboolean    *was_restarted)
+{
+        *was_restarted = TRUE;
+}
+
+static void
+do_http_1_1_required_test (Test *test, gconstpointer data)
+{
+        SoupMessage *msg;
+        GBytes *response;
+        gboolean was_restarted = FALSE;
+        GError *error = NULL;
+
+        SOUP_TEST_SKIP_IF_NO_APACHE;
+
+        msg = soup_message_new (SOUP_METHOD_GET, "https://127.0.0.1:47525/client-cert");
+        soup_message_set_force_http_version (msg, SOUP_HTTP_2_0);
+        g_signal_connect (msg, "restarted",
+                          G_CALLBACK (message_restarted), &was_restarted);
+        response = soup_test_session_async_send (test->session, msg, NULL, &error);
+        g_assert_no_error (error);
+        g_assert_cmpuint (soup_message_get_status (msg), ==, 403);
+        g_assert_true (was_restarted);
+        g_assert_nonnull (response);
+        g_bytes_unref (response);
+        g_object_unref (msg);
+
+        while (g_main_context_pending (NULL))
+                g_main_context_iteration (NULL, FALSE);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -1025,6 +1054,8 @@ main (int argc, char **argv)
                 return 1;
         }
 
+        apache_init ();
+
         g_test_add ("/http2/basic/async", Test, NULL,
                     setup_session,
                     do_basic_async_test,
@@ -1125,6 +1156,10 @@ main (int argc, char **argv)
                     setup_session,
                     do_timeout_test,
                     teardown_session);
+        g_test_add ("/http2/http-1-1-required", Test, NULL,
+                    setup_session,
+                    do_http_1_1_required_test,
+                    teardown_session);
 
 	ret = g_test_run ();
 
